#!/usr/bin/env python3

import numpy as np
import click
from mdkits.util import encapsulated_ase, os_operation


def ave_cube_data(cube_data, range):
	mask = (cube_data[:,0] >= range[0]) & (cube_data[:,0] <=range[1])
	bulk_cube_data = cube_data[mask]
	ave_cube_data = np.mean(bulk_cube_data[:,1])
	return ave_cube_data


@click.command(name='cube')
@click.argument('filename', type=click.Path(exists=True), default=os_operation.default_file_name('*.cube', last=True))
@click.argument('axis', type=click.Choice(['x','y','z', '0','1','2']), default='z')
@click.option('-b', '--bulk_range', type=(float, float), help='parameter to calculate mean value of bulk', default=None)
@click.option('-o', type=str, help='output file name, default is "cube.out"', default='cube.out', show_default=True)
def main(filename, axis, bulk_range, o):
	"""
	analysis cube file
	"""

	cube_data = encapsulated_ase.ave_cube(filename, axis)

	## if bulk range is exit, out put a difference of cube_data
	if bulk_range is not None:
		bulk_cube_data = ave_cube_data(cube_data, bulk_range)
		print(bulk_cube_data)
		np.savetxt(o, cube_data, header=f'Z\tcube_data\t area average is: {bulk_cube_data}')
	
	else:
		np.savetxt(o, cube_data, header='Z\tcube_data')

	


if __name__ == '__main__':
	main()


import click
from mdkits.build_cli import build_cli
from mdkits.dft_cli import dft_cli
from mdkits.dfpt_cli import dfpt_cli
from mdkits.md_cli import md_cli
from mdkits.cli import (
    convert,
    extract,
    data,
    plot,
)


@click.group(name='cli')
@click.pass_context
@click.version_option()
def cli(ctx):
    """kits for md or dft"""
    pass


cli.add_command(md_cli.cli)
cli.add_command(build_cli.cli_build)
cli.add_command(dft_cli.main)
cli.add_command(dfpt_cli.main)
cli.add_command(convert.main)
cli.add_command(extract.main)
cli.add_command(data.main)
cli.add_command(plot.main)


if __name__ == '__main__':
    cli()


#!/usr/bin/env python3

# extract final structure form pos.xyz file

import os
import click
from mdkits.util import os_operation, arg_type, out_err
import MDAnalysis
from MDAnalysis import Universe


def write_to_xyz(u, frames, o, select, cut=None):
    ag = u.select_atoms(select)
    with MDAnalysis.Writer(o, ag.atoms.n_atoms, format='XYZ') as w:
        for ts in u.trajectory:
            if ts.frame in frames:
                w.write(ag)
    if cut:
        with open(o, 'r') as fi, open(o+'t', 'w') as fo:
            for i, line in enumerate(fi):
                if i >= cut:
                    fo.write(line)
        os.replace(o+'t', o)


def write_to_xyz_s(u, frames, select, cut=None):
    index = 0
    ag = u.select_atoms(select)
    if select:
        dir = f'./coord/{"_".join(select.split())}'
    else:
        dir = './coord/all'
    for ts in u.trajectory:
        if ts.frame in frames:
            o = f'{dir}/coord_{index:03d}.xyz'
            with MDAnalysis.Writer(o, ag.atoms.n_atoms, format='XYZ') as w:
                w.write(ag)
                index += 1
            if cut:
                with open(o, 'r') as fi, open(o+'t', 'w') as fo:
                    for i, line in enumerate(fi):
                        if i >= cut:
                            fo.write(line)
                os.replace(o+'t', o)

@click.command(name='extract')
@click.argument('input_file_name', type=click.Path(exists=True), default=os_operation.default_file_name('*-pos-1.xyz', last=True))
@click.option('-r', type=arg_type.FrameRange, help='frame range to slice', default='-1', show_default=True)
@click.option('-c', help='output a coord.xyz', is_flag=True)
@click.option("--select", type=str, help="select atoms to extract", default="all", show_default=True)
def main(input_file_name, r, c, select):
    """
    extract frames in trajectory file
    """

    u = Universe(input_file_name)
    if len(r) == 1:
        print(f"frame range slice is {r}")
        group = u.trajectory[r]
    else:
        print(f"frame range slice is {slice(*r)}")
        group = u.trajectory[slice(*r)]
    click.echo(f"total frames is {len(u.trajectory)}")
    frames = [ts.frame for ts in group]

    if c:
        cut = 2
    else:
        cut = None

    if len(r) == 3 and r[-1] is not None:
        if select:
            dir = f'./coord/{"_".join(select.split())}'
        else:
            dir = './coord/all'
        if not os.path.exists(dir):
            os.makedirs(dir)
        else:
            import shutil
            shutil.rmtree(dir)
            os.makedirs(dir)
        write_to_xyz_s(u, frames, select, cut=cut)
        click.echo(os.path.abspath(dir))
    else:
        o = f"{os.path.basename(u.filename).split('.')[0]}_{'_'.join([str(i) for i in r])}_{'_'.join(select.split()) if select else 'all'}.xyz"
        write_to_xyz(u, frames, o, select, cut=cut)
        out_err.path_output(o)


if __name__ == '__main__':
    main()



import click
from mdkits.util import arg_type, encapsulated_ase
import numpy as np


@click.command(name='dope')
@click.argument('atoms', type=arg_type.Structure)
@click.option('--group', type=str, required=True, help='group to dope')
@click.option('--element', type=(str, int), required=True, help='element and number of element to dope', multiple=True)
@click.option('--seed', type=int, default=20000203, help='random seed')
def main(atoms, group, element, seed):
    """Dope atoms in the structure"""
    np.random.seed(seed)

    u = encapsulated_ase.atoms_to_u(atoms)
    u.dimensions = atoms.cell.cellpar()

    for elementi, times in element:
        atomgroup = u.select_atoms(f"{group}")
        if atomgroup.n_atoms == 0:
            raise ValueError(f"No atoms found for {group} in doping {elementi}")
        if times > atomgroup.n_atoms:
            raise ValueError(f"Cannot dope {times} atoms, only {atomgroup.n_atoms} atoms available in group {group}")
        selected_indices = np.random.choice(atomgroup.indices, size=times, replace=False)
        for idx in selected_indices:
            u.atoms[idx].element = elementi

    element, count = np.unique(u.select_atoms("all").elements, return_counts=True)
    o = f"{atoms.filename.split('.')[0]}_{'-'.join([f'{e}{c}' for e, c in zip(element, count)])}_doped.cif"

    encapsulated_ase.u_to_cif(u, o)


#!/usr/bin/env python3

import numpy as np
import click, math
from MDAnalysis import Universe
from MDAnalysis.analysis.base import AnalysisBase
from mdkits.util import (
    arg_type,
    numpy_geo,
    encapsulated_mda,
    os_operation,
)
from .setting import common_setting
import warnings, sys
warnings.filterwarnings("ignore")


class Density_distribution(AnalysisBase):
    def __init__(self, filename, cell, o, element, atomic_mass, update_water, distance_judg, angle_judg, surface, dt=0.001, bin_size=0.2, return_index=False):
        u = Universe(filename)
        u.trajectory.ts.dt = dt
        u.dimensions = cell

        self.u = u
        self.atomgroup = u.select_atoms("all")
        self.bin_size = bin_size
        self.frame_count = 0
        self.return_index = return_index
        self.surface = surface
        self.atomic_mass = atomic_mass
        self.update_water = update_water
        self.element = element

        if element is None and update_water is False:
            sys.exit("Please specify the element to analysis or use --update-water option")

        if self.update_water:
            self.distance_judg = distance_judg
            self.angle_judg = angle_judg

        if o == 'density_{element}.dat':
            if self.update_water:
                self.o = "density_water.dat"
            else:
                self.o = f"density_{element.replace(' ', '_')}.dat"
        else:
            self.o = o

        if surface is not None:
            self.surface_group = self.atomgroup.select_atoms(f"{surface}")
            if self.surface_group.n_atoms == 0:
                sys.exit("Please specify the correct surface group")
        else:
            self.surface_group = False

        super(Density_distribution, self).__init__(self.atomgroup.universe.trajectory, verbose=True)

    def _prepare(self):
        self.bin_num = int(self.u.dimensions[2] / self.bin_size) + 2
        self.density_distribution = np.zeros(self.bin_num, dtype=np.float64)
        if self.surface_group:
            self.surface_pos = np.zeros(2)

    def _append(self, z):
        bins = np.floor(z / self.bin_size).astype(int) + 1
        np.add.at(self.density_distribution, bins, 1)


    def _single_frame(self):
        if self.update_water:
            o_group = self.atomgroup.select_atoms("name O")
            h_group = self.atomgroup.select_atoms("name H")

            o, oh1, oh2 = encapsulated_mda.update_water(self, o_group, h_group, distance_judg=self.distance_judg, angle_judg=self.angle_judg, return_index=self.return_index)

            self._append(o.positions[:, 2])

        else:
            group = self.atomgroup.select_atoms(f"{self.element}", updating=True)
            self._append(group.positions[:, 2])

        if self.surface_group:
            surface = numpy_geo.find_surface(self.surface_group.positions[:, 2])
            self.surface_pos[0] += surface[0]
            self.surface_pos[1] += surface[1]

        self.frame_count += 1

    def _conclude(self):
        if self.frame_count > 0:
            V = self.u.dimensions[0] * self.u.dimensions[1] * math.sin(math.radians(180 - self.u.dimensions[-1])) * self.bin_size

            if self.atomic_mass:
                density_distribution = (self.density_distribution * self.atomic_mass * 1.660539 / V) / self.frame_count
            else:
                density_distribution = (self.density_distribution * (10000/6.02) / V) / self.frame_count

            bins_z = np.arange(len(self.density_distribution)) * self.bin_size + self.bin_size / 2

            if self.surface:
                lower_z = self.surface_pos[0] / self.frame_count
                if self.surface_pos[1] == 0:
                    upper_z = np.inf
                else:
                    upper_z = self.surface_pos[1] / self.frame_count

                mask = (bins_z >= lower_z) & (bins_z <= upper_z)
                filtered_bins_z = bins_z[mask] - lower_z
                filtered_density_distribution = density_distribution[mask]

                conbined_data = np.column_stack((filtered_bins_z, filtered_density_distribution))
            else:
                conbined_data = np.column_stack((bins_z, density_distribution))

            np.savetxt(self.o, conbined_data, header="Z\tdensity", fmt='%.5f', delimiter='\t')

@click.command(name='density', help="analysis density or concentration of element in a trajectory file")
@common_setting
@click.option('--group', type=str, help='group to analysis')
@click.option('--atomic_mass', type=float, help='output density unit (g/cm3), should give atomic mass of element, else is concentration unit (mol/L)')
@click.option('-o', type=str, help='output file name', default='density_{element}.dat', show_default=True)
def main(filename, cell, o, group, atomic_mass, update_water, distance, angle, surface, r):
    """
    analysis density or concentration of element in a trajectory file
    """

    density_dist = Density_distribution(filename, cell, o=o, distance_judg=distance, angle_judg=angle, element=group, atomic_mass=atomic_mass, update_water=update_water, surface=surface)

    if r is not None:
        if len(r) == 2:
            density_dist.run(start=r[0], stop=r[1])
        elif len(r) == 3:
            density_dist.run(start=r[0], stop=r[1], step=r[2])
    else:
        density_dist.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import click, os
from ase.build import bulk
import numpy as np
from mdkits.util import out_err


@click.command(name='bulk')
@click.argument('symbol', type=str)
@click.argument('cs', type=click.Choice(['sc', 'fcc', 'bcc', 'tetragonal', 'bct', 'hcp', 'rhombohedral', 'orthorhombic', 'mcl', 'diamond', 'zincblende', 'rocksalt', 'cesiumchloride', 'fluorite', 'wurtzite']))
@click.option('-a', type=float, help='designate lattice constant a')
@click.option('-b', type=float, help='designate lattice constant b. if only a and b is given, b will be interpreted as c instead')
@click.option('-c', type=float, help='designate lattice constant c')
@click.option('--alpha', type=float, help='angle in degrees for rhombohedral lattice')
@click.option('--covera', type=float, help='c/a ratio used for hcp. Default is ideal ratio: sqrt(8/3)', default=np.sqrt(8/3), show_default=True)
@click.option('-u', type=float, help='internal coordinate for Wurtzite structure')
@click.option('--orth', is_flag=True, help='construct orthorhombic unit cell instead of primitive cell which is the default')
@click.option('--cubic', is_flag=True, help='construct cubic unit cell if possible')
def main(symbol, cs, a, b, c, alpha, covera, u, orth, cubic):
    """
    build a bulk structure
    """

    #if args.primitve:
    #    a = args.a * 0.7071 * 2
    #else:
    #    a = args.a
    atoms = bulk(symbol, cs, a=a, b=b, c=c, alpha=alpha, covera=covera, u=u, orthorhombic=orth, cubic=cubic)

    o = f"{symbol}_{cs}.cif"
    atoms.write(o, format='cif')
    out_err.path_output(o)


if __name__ == '__main__':
    main()




import click
from mdkits.dfpt_cli import (
    elastic,
)


@click.group(name='dfpt')
@click.pass_context
def main(ctx):
    """kits for dfpt analysis"""
    pass


main.add_command(elastic.main)

if __name__ == '__main__':
    main()


"""
output and error for cli
"""

import numpy as np
import sys, os


def cell_output(atoms):
    cell = atoms.cell.cellpar()
    if not hasattr(atoms, "name"):
        atoms.name = "present"
    print(f"{atoms.name} cell: x = {cell[0]}, y = {cell[1]}, z = {cell[2]}, a = {cell[3]}\u00B0, b = {cell[4]}\u00B0, c = {cell[5]}\u00B0")


def path_output(file: str):
    env_var_name = 'ssh_name'
    file_path = os.path.abspath(file)
    if os.environ.get(env_var_name):
        ssh_name = os.environ.get(env_var_name)
        file_path = f"{ssh_name}:{file_path}"
    print(file_path)

def check_cell(atoms, cell=None):
    if cell is not None:
        atoms.set_cell(cell)
        cell_output(atoms)
    elif not np.array_equal(atoms.cell.cellpar(), np.array([0., 0., 0., 90., 90., 90.])):
        cell_output(atoms)
    elif np.array_equal(atoms.cell.cellpar(), np.array([0., 0., 0., 90., 90., 90.])) and cell is not None:
        atoms.set_cell(cell)
        cell_output(atoms)
    else:
        raise ValueError("can't parse cell please use --cell set cell")



import click
import MDAnalysis as mda
import MDAnalysis.analysis.msd as msd
import numpy as np
from mdkits.util import arg_type


@click.command(name="msd")
@click.argument("filename", type=click.Path(exists=True))
@click.argument('type', type=click.Choice(['xyz', 'xy', 'yz', 'xz', 'x', 'y', 'z']))
@click.argument("group", type=str)
@click.option('-r', type=arg_type.FrameRange, help='range of frame to analysis')
def main(filename, type, group, r):
    """analysis msd along the given axis"""
    u = mda.Universe(filename)
    MSD = msd.EinsteinMSD(u, select=group, msd_type=type, fft=True)
    if r is not None:
        if len(r) == 2:
            MSD.run(start=r[0], stop=r[1], verbose=True)
        elif len(r) == 3:
            MSD.run(start=r[0], stop=r[1], step=r[2], verbose=True)
    else:
        MSD.run(verbose=True)

    data = np.arange(1, MSD.n_frames + 1).reshape(-1, 1)
    s = "_"
    name = f"{s.join(group.split(' '))}"
    o = f"msd_{type}_{name}.dat"
    header = ''
    msd_cols = []
    for i in range(MSD.n_particles):
        msd_cols.append(MSD.results.msds_by_particle[:, i].reshape(-1, 1))
        header += name + f"_{i}\t"
    msd_array = np.concatenate(msd_cols, axis=1)
    mean_col = np.mean(msd_array, axis=1, keepdims=True)
    data = np.concatenate((data, mean_col, msd_array), axis=1)
    header = "frame\tmean\t" + header

    np.savetxt(o, data, fmt="%.5f", delimiter="\t", header=header)



if __name__ == '__main__':
    main()


import click, json
from pymatgen.analysis import elasticity
from mdkits.util import arg_type, encapsulated_ase
import numpy as np


def voigt_to_tensor(C_voigt):

    voigt_map = {
        0: (0, 0),
        1: (1, 1),
        2: (2, 2),
        3: (1, 2),
        4: (0, 2),
        5: (0, 1),
    }

    C_tensor = np.zeros((3, 3, 3, 3))

    for alpha in range(6):
        for beta in range(6):
            i, j = voigt_map[alpha]
            k, l = voigt_map[beta]

            value = C_voigt[alpha, beta]

            for ii, jj in [(i, j), (j, i)]:
                for kk, ll in [(k, l), (l, k)]:
                    C_tensor[ii, jj, kk, ll] = value
                    C_tensor[kk, ll, ii, jj] = value

    return C_tensor


@click.command(name='elastic')
@click.argument('structure', type=arg_type.Structure)
@click.argument('ela_data', type=click.Path(exists=True))
@click.option('--cell', type=arg_type.Cell, help='set cell, a list of lattice: --cell x,y,z or x,y,z,a,b,c')
def main(structure, ela_data, cell):
    if cell:
        structure.set_cell(cell)

    stru = encapsulated_ase.atoms_to_pymatgen_structure(structure)

    data = np.loadtxt(ela_data)
    tensor = voigt_to_tensor(data*100)

    elasitic_tensor = elasticity.ElasticTensor(tensor)
    elasitic_properties = elasitic_tensor.get_structure_property_dict(stru)
    elasitic_properties.pop('structure')

    text = json.dumps(elasitic_properties, indent=4)

    print(text)
    with open('elasitic_properties.json', 'w') as f:
        f.write(text)


"""
filename: encapsulated_ase.py
function: some encapsulated ase method
"""


from ase.io import iread, read
import io
import numpy as np
from ase.io.cube import read_cube_data
import MDAnalysis


def wrap_to_cell(chunk, cell, name, big=False):
    """
    function: encapsulated ase.Atom.wrap method
    parameter:
        chunk: a list of frame of structure
        cell: a list of cell parameter
        name: path of temp file to write
        big: make 3X3X1 supercell
    return:
        write into a temp file
    """
    # wrap a chunk of .xyz sturcture in a cell, and write wraped structure to file
    if len(chunk) > 1:
        atoms = iread(io.StringIO('\n'.join(chunk)), format='xyz')
    else:
        atoms = iread(chunk, format='xyz')
    with open(name, 'w') as f:
        f.write('')
    for atom in atoms:
        atom.set_cell(cell)
        atom.set_pbc(True)
        atom.wrap()
        if big:
            from ase.build import make_supercell
            P = [ [3, 0, 0], [0, 3, 0], [0, 0, 1] ]
            atom = make_supercell(atom, P)
        atom.write(f'{name}', append='a', format='xyz')


def rdf(chunk, cell, bin_size, name, parallel=True):
    """
    not used
    """
    from ase.geometry.analysis import Analysis
    from math import ceil
    n = ceil(cell[2] / bin_size)
    if parallel:
        atoms = iread(io.StringIO('\n'.join(chunk)), format='xyz')
    else:
        atoms = iread(chunk, format='xyz')
    atom_list = []
    for atom in atoms:
        atom.set_cell(cell)
        atom.set_pbc(True)
        atom_list.append(atom)
    print(len(atom_list))
    ana = Analysis(atom_list)
    rdf = ana.get_rdf(3.5, n, elements=['O'])
    rdf_array = np.vstack(rdf)
    rdf_array = np.mean(rdf_array, axis=0)
    print(name)
    np.save(name, rdf_array)
    ana.clear_cache()


def ave_cube(filepath, axis):
    """
    function: average hartree file in z_coordinate
    parameter:
        filepath: hartree cube file path
        axis: average axis, can be 'x','y','z' or 0,1,2
    return:
        z_cube_data: a list of cube data alone z axes
        z_coordinates: a list of coordinates of z axes
    """
    if isinstance(axis, str):
        mapping = {'x': 0, 'y': 1, 'z': 2}
        key = axis.lower()
        if key not in mapping:
            raise ValueError(f"axis string must be one of 'x','y','z', got {axis}")
        axis_idx = mapping[key]
    else:
        axis_idx = int(axis)
    
    # read data from filepath
    data, atoms = read_cube_data(filepath)

    ndim = data.ndim
    if axis_idx < 0:
        axis_idx = ndim + axis_idx
    if not (0 <= axis_idx < ndim):
        raise ValueError(f"axis must be between 0 and {ndim-1}, got {axis_idx}")

    npoints = data.shape[axis_idx]
    step_size = atoms.cell.cellpar()[axis_idx] / ( npoints - 1 )

    z_coordinates = [i * step_size for i in range(npoints)]

    other_axes = tuple(i for i in range(ndim) if i != axis_idx)
    if other_axes:
        averaged = data.mean(axis=other_axes)
    else:
        averaged = data.copy()
    return np.column_stack((z_coordinates, averaged))


def atoms_read_with_cell(file_name, cell=None, coord_mode=False, default_cell=np.array([0., 0., 0., 90., 90., 90.])):
    """
    function: read structure file return a Atoms object with cell
    parameter:
        file_name: structure file name
        cell: cell list
        coord_mode: if file name is coord format
        default_cell: default cell of Atoms object
    return:
        atoms: Atoms object with cell
    """
    import sys, os
    from . import cp2k_input_parsing, structure_parsing

    if coord_mode:
        xyz = structure_parsing.coord_to_xyz(file_name)
        atoms = read(io.StringIO('\n'.join(xyz)), format='xyz')
    else:
        atoms = jread(file_name)

    if cell == None:
        cp2k_cell_file_name = ( 'cell.inc', 'setup.inp', 'input.inp' )
        existing_cell_file = [ file for file in cp2k_cell_file_name if os.path.isfile(file) ]
        if existing_cell_file:
            cell = cp2k_input_parsing.parse_cell(existing_cell_file[0])

    if np.array_equal(atoms.cell.cellpar(), default_cell):
        if cell != None:
            atoms.set_cell(cell)
            print('cell: \n' + ', '.join(map(str, atoms.cell.cellpar())))
            return atoms
        else:
            print("no cell")
            sys.exit(1)
    print('cell: \n' + ', '.join(map(str, atoms.cell.cellpar())))
    return atoms


def jread(filepath):
    from ase.io import read
    suffix = filepath.split('.')[-1]
    if suffix == 'data':
        atom = read(filepath, format='lammps-data', atom_style='atomic')
    else:
        atom = read(filepath)

    return atom


def atoms_to_u(atoms):
    virtual_file = io.StringIO()
    cell = atoms.cell.cellpar()
    atoms.write(virtual_file, format='xyz')

    u = MDAnalysis.Universe(virtual_file, format='xyz')
    u.dimensions = cell

    return u

def atoms_to_pymatgen_structure(atoms):
    import os
    temp_file = './temp.cif'
    atoms.write(temp_file, format='cif')

    from pymatgen.io import cif
    structure = cif.CifParser(temp_file).get_structures(primitive=False)[0]

    if os.path.exists(temp_file):
        os.remove(temp_file)

    return structure

def u_to_cif(u, filename):
    import os
    temp_file = './temp.xyz'
    u.select_atoms("all").write(temp_file, format='xyz')


    atoms = read(temp_file)

    if os.path.exists(temp_file):
        os.remove(temp_file)

    atoms.set_cell(u.dimensions)
    atoms.write(filename, format='cif')



from pymatgen.io.cp2k import outputs
import click
from mdkits.util import arg_type
from pymatgen.electronic_structure.core import Spin
import pandas as pd
import numpy as np


def calculate_gaussian_dos(E_grid, eigenvalues, weights, sigma):
    """
    使用高斯函数对离散能级进行展宽计算 DOS。
    利用 NumPy 广播机制进行高效计算。
    """
    # E_grid shape: (N_grid,) -> 变为列向量 (N_grid, 1)
    # eigenvalues shape: (N_eigen,) -> 变为行向量 (1, N_eigen)
    # delta_E shape: (N_grid, N_eigen)，包含了每个网格点到每个本征值的距离
    delta_E = E_grid[:, np.newaxis] - eigenvalues[np.newaxis, :]
    
    # 计算高斯项
    prefactor = 1.0 / (sigma * np.sqrt(2 * np.pi))
    gaussians = prefactor * np.exp(-0.5 * (delta_E / sigma)**2)
    
    # 对所有本征值求和，并乘上权重
    # axis=1 表示沿着本征值的方向求和，最终得到 (N_grid,) 的数组
    dos_values = np.sum(gaussians * weights[np.newaxis, :], axis=1)
    
    return dos_values



@click.command(name='pdos')
@click.option('-i', type=arg_type.FileList, default='*-k*.*', help='pdos files generated by CP2K', show_default=True)
@click.option('-s', type=float, default=0.2, help='energy step length (eV)', show_default=True)
@click.option('--no-reference', is_flag=True, help='if specified, reference energy will not be shifted to fermi level')
@click.option('-o', type=str, help='output file name')
def main(i, s, no_reference, o):
    filename = i
    total_dict = {}
    for file in filename:
        dos = outputs.parse_pdos(file, total=True)
        
        pdos = dos[0]
        tdos = dos[1]
        element = list(pdos.keys())[0]
        data_dict = {}
        steplen = s
        fermi = pdos[element][list(pdos[element].keys())[0]].efermi
        gap =  pdos[element][list(pdos[element].keys())[0]].get_gap()
        sigma = 0.02

        energies = pdos[element][list(pdos[element].keys())[0]].energies
        n_min = np.floor((energies.min() - fermi) / steplen)
        n_max = np.ceil((energies.max() - fermi) / steplen)
        integers = np.arange(n_min, n_max + 1)
        bins = fermi + integers * steplen
        #e_min = energies.min() - 3 * sigma
        #e_max = energies.max() + 3 * sigma
        #n_step_left = np.ceil((fermi - e_min) / steplen).astype(int)
        #n_step_right = np.ceil((e_max - fermi) / steplen).astype(int)
        #indices = np.arange(-n_step_left, n_step_right + 1)
        #bins = fermi + indices * steplen
        #bins = int((energies[-1]-energies[0])/steplen)


        dos, ener = np.histogram(energies, bins=bins, weights=tdos.get_densities(Spin.up), range=(energies[0], energies[-1]))
        #dos = calculate_gaussian_dos(bins + steplen/2, energies, tdos.get_densities(Spin.up), sigma)
        #data_dict['energy'] = bins + steplen/2
        data_dict['energy'] = ener[:-1] + steplen/2


        df = pd.DataFrame(data_dict)

        df['total'] = 0
        for orb_key in pdos[element].keys():
            data_obj = pdos[element][orb_key]
            weights = data_obj.get_densities(Spin.up)
            dos, ener = np.histogram(energies, bins=bins, weights=weights, range=(energies[0], energies[-1]))
            #dos = calculate_gaussian_dos(bins + steplen/2, energies, weights, sigma)
            df[orb_key] = dos/steplen
            df['total'] += dos/steplen


        if Spin.down in pdos[element][list(pdos[element].keys())[0]].densities:
            df['total_down'] = 0
            for orb_key in pdos[element].keys():
                data_obj = pdos[element][orb_key]
                weights = data_obj.get_densities(Spin.down)
                dos, ener = np.histogram(energies, bins=bins, weights=weights, range=(energies[0], energies[-1]))
                #dos = calculate_gaussian_dos(bins + steplen/2, energies, weights, sigma)
                df[f'{orb_key}_down'] = dos/steplen
                df['total_down'] += dos/steplen

        if o is None:
            out_filename = f"{element}.pdos"
        else:
            out_filename = o

        total_dict[element] = df

        if not no_reference:
            df['energy'] = df['energy'] - fermi

        with open(out_filename, 'w') as f:
            header = '#' + ' '.join([str(key) for key in df.columns]) + '\n'
            f.write(header)
            df.to_csv(f, sep='\t', index=False, header=False)


    energy = list(total_dict.values())[0]['energy']

    total_dos = 0
    for element, df in total_dict.items():
        total_dos += df['total']

    total_df = pd.DataFrame({'energy': energy, 'total': total_dos})

    
    with open('total.pdos', 'w') as f:
        header = '# energy total_dos' + '\n' + f'# Fermi Level: {fermi}' + f'gap: {gap}' + '\n'
        f.write(header)
        total_df.to_csv(f, sep='\t', index=False, header=False)


import click
from MDAnalysis import Universe
from MDAnalysis.analysis.base import AnalysisBase
import MDAnalysis
import sys
from mdkits.util import numpy_geo, encapsulated_mda, arg_type
import numpy as np
from .setting import common_setting


class Angle_distribution(AnalysisBase):
    def __init__(self, filename, cell, water_height, update_water, surface, distance_judg=None, angle_judg=(None, None), dt=0.001, bin_size=5):
        u = Universe(filename)
        u.trajectory.ts.dt = dt
        u.dimensions = cell

        self.u = u
        self.atomgroup = u.select_atoms("all")
        self.bin_size = bin_size
        self.frame_count = 0
        self.surface = surface
        self.update_water = update_water
        self.mid_z = u.dimensions[2]/2

        self.normal_up = np.array([0, 0, 1])
        self.normal_down = np.array([0, 0, -1])
        self.total_angle = 180

        if water_height is None:
            sys.exit("Please specify the water height")
        else:
            self.water_height = water_height

        if self.update_water:
            self.distance_judg = distance_judg
            self.angle_judg = angle_judg

        if surface is not None:
            self.surface_group = self.atomgroup.select_atoms(f"{surface}")
            if self.surface_group.n_atoms == 0:
                sys.exit("Please specify the correct surface group")
        else:
            sys.exit("Please specify a surface group")

        super(Angle_distribution, self).__init__(self.atomgroup.universe.trajectory, verbose=True)
    
    def _prepare(self):
        self.bin_num = int(self.total_angle / self.bin_size) + 2
        self.angle_w_distribution = np.zeros(self.bin_num, dtype=np.float64)
        self.angle_oh_distribution = np.zeros(self.bin_num, dtype=np.float64)

    def _append(self, angle, anglew=True):
        bins = np.floor(angle / self.bin_size).astype(int) + 1

        if anglew:
            bins = bins[bins < len(self.angle_w_distribution)]
            np.add.at(self.angle_w_distribution, bins, 1)
        else:
            bins = bins[bins < len(self.angle_oh_distribution)]
            np.add.at(self.angle_oh_distribution, bins, 1)


    def _single_frame(self):
        surface = numpy_geo.find_surface(self.surface_group.positions[:, 2])

        if surface[1] == 0:
            o_group = self.atomgroup.select_atoms(f"name O and prop z < {surface[0]+self.water_height}", updating=True)
        else:
            o_group = self.atomgroup.select_atoms(f"name O and (prop z < {surface[0]+self.water_height} or prop z > {surface[1]-self.water_height})", updating=True)
        
        h_group = self.atomgroup.select_atoms("name H")

        if self.update_water:
            o, oh1, oh2 = encapsulated_mda.update_water(self, o_group=o_group, h_group=h_group, distance_judg=self.distance_judg, angle_judg=self.angle_judg, return_index=False)
        else:
            o = o_group
            oh1 = self.atomgroup[o_group.indices + 1]
            oh2 = self.atomgroup[o_group.indices + 2]

        vec1 = MDAnalysis.lib.distances.minimize_vectors(oh1.positions - o.positions, self.u.dimensions)
        vec2 = MDAnalysis.lib.distances.minimize_vectors(oh2.positions - o.positions, self.u.dimensions)

        bisector = numpy_geo.vector_between_two_vector(vec1, vec2)

        angle_vec1 = np.hstack((numpy_geo.vector_vector_angle(vec1[o.positions[:, 2] < self.mid_z], self.normal_up), numpy_geo.vector_vector_angle(vec1[o.positions[:, 2] > self.mid_z], self.normal_down)))

        angle_vec2 = np.hstack((numpy_geo.vector_vector_angle(vec2[o.positions[:, 2] < self.mid_z], self.normal_up), numpy_geo.vector_vector_angle(vec2[o.positions[:, 2] > self.mid_z], self.normal_down)))

        angle_bisector = np.hstack((numpy_geo.vector_vector_angle(bisector[o.positions[:, 2] < self.mid_z], self.normal_up), numpy_geo.vector_vector_angle(bisector[o.positions[:, 2] > self.mid_z], self.normal_down)))

        self._append(angle_vec1, anglew=False)
        self._append(angle_vec2, anglew=False)
        self._append(angle_bisector)

        self.frame_count += 1

    def _conclude(self):
        if self.frame_count > 0:
            average_angle_w = self.angle_w_distribution / self.frame_count
            average_angle_oh = self.angle_oh_distribution / (self.frame_count*2)
            bins_z = np.arange(len(average_angle_w)) * self.bin_size + self.bin_size / 2
            conbined_data = np.column_stack((bins_z, average_angle_w, average_angle_oh))
            np.savetxt("angle_distribution.dat", conbined_data, header="angle\tw_suf_dist\toh_suf_dist", fmt='%.5f', delimiter='\t')


@click.command(name="angle", help="analysis angle between normal vectors and OH vector or bisector")
@common_setting
@click.option("--water_height", type=float, help="water height from surface")
def main(filename, cell, water_height, update_water, distance, angle, surface, r):
    """analysis angle between normal vectors and OH vector or bisector"""
    a = Angle_distribution(filename, cell, water_height, update_water, distance_judg=distance, angle_judg=angle, surface=surface)
    if r is not None:
        if len(r) == 2:
            a.run(start=r[0], stop=r[1])
        elif len(r) == 3:
            a.run(start=r[0], stop=r[1], step=r[2])
    else:
        a.run()


if __name__ == "__main__":
    main()


import click, os
from ase.io import read
import numpy as np
from ase.collections import g2
from mdkits.util import os_operation, cp2k_input_parsing, out_err


class CellType(click.ParamType):
    name = "cell type"

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            cell = [float(x) for x in value.split(',')]

            if len(cell) == 3:
                cell += [90, 90, 90]
                return cell
            elif len(cell) == 6:
                return cell
            else:
                self.fail(f"{value} is not a valid cell parameter", param, ctx)


class FrameRangeType(click.ParamType):
    name = "frame range"
    def convert(self, value, param, ctx):
        if isinstance(value, str):
            parts = value.split(':')

            range_list = [int(x) if x else None for x in parts]

            if len(range_list) > 0 and len(range_list) <= 3:
                return range_list
            else:
                self.fail(f"{value} is not a valid frame range", param, ctx)


class StructureType(click.ParamType):
    name = "structure type"
    def convert(self, value, param, ctx):
        no_cell=np.array([0., 0., 0., 90., 90., 90.])
        if isinstance(value, str):
            if os.path.exists(value):
                try:
                    atoms = read(value)
                except:
                    self.fail(f"{value} is not a valid structure file", param, ctx)

                if np.array_equal(atoms.cell.cellpar(), no_cell):
                    cell = cp2k_input_parsing.parse_cell()
                    atoms.set_cell(cell)

                atoms.filename = value.replace('./', '').replace('.\\', '').split('/')[-1]
                return atoms
            else:
                self.fail(f"{value} is not exists", param, ctx)


class MoleculeType(click.Choice):
    name = "mocular type"
    def __init__(self):
        g2.names.append(click.Path(exists=True))
        super().__init__(choices=tuple(g2.names))


class AdsSiteType(click.Choice):
    name = "adsorption site"
    def __init__(self):
        site = ['ontop', 'hollow','fcc', 'hcp', 'bridge', 'shortbridge', 'longbridge']
        super().__init__(choices=tuple(site))


class FileListType(click.ParamType):
    name = "file list"
    def convert(self, value, param, ctx):
        if isinstance(value, str):
            value = value.replace('`', '') # for window adaptation
            import glob
            file_list = glob.glob(value)
            if file_list:
                return file_list
            else:
                self.fail(f"No files match {value}", param, ctx)



Cell = CellType()
FrameRange = FrameRangeType()
Molecule = MoleculeType()
AdsSite = AdsSiteType()
Structure = StructureType()
FileList = FileListType()


import click
from MDAnalysis import Universe
from mdkits.util import encapsulated_ase, os_operation


@click.command(name='fix')
@click.argument('filename', type=click.Path(exists=True), default='./coord.xyz')
@click.argument('group', type=str)
@click.option('-o', type=str, help='output file name, default is "fix.inc"', default='fix.inc', show_default=True)
def main(filename, group, o):
    """
    generate fix.inc file for cp2k
    """
    
    with open(filename, 'r', encoding='utf-8') as fh:
        lines = fh.read().splitlines()
    first = lines[0].strip() if lines else ''
    if not first.isdigit():
        atoms_number = sum(1 for l in lines if l.strip())
        s = '\n'.join([str(atoms_number), ''] + lines) + '\n'
        import io
        virtual_file = io.StringIO(s)
        atoms = Universe(virtual_file, format='xyz').select_atoms(group)
    else:
        atoms = Universe(filename).select_atoms(group)
    indices = atoms.indices + 1
    fix_number = len(indices)
    fix_output = f"{fix_number} atoms have been fixed."
    
    arr = sorted(set(int(i) for i in indices))
    if not arr:
        list_str = ""
    else:
        ranges = []
        start = prev = arr[0]
        for x in arr[1:]:
            if x == prev + 1:
                prev = x
            else:
                if start == prev:
                    ranges.append(str(start))
                else:
                    ranges.append(f"{start}..{prev}")
                start = prev = x
        
        if start == prev:
            ranges.append(str(start))
        else:
            ranges.append(f"{start}..{prev}")
        list_str = "# " + group + "\n" "# " + fix_output + "\n" + "LIST " + " ".join(ranges)

    print(list_str)
    with open(o, 'w') as f:
        f.write(list_str + '\n')